Require Export PropiL.


(** * Classical Natural Deduction *)

Inductive ndc : context -> form -> Prop :=
  | ndcA A s:      s el A -> ndc A s
  | ndcC A s:      ndc (Not s :: A) Fal -> ndc A s
  | ndcII A s t:   ndc (s::A) t -> ndc A (Imp s t)
  | ndcIE A s t:   ndc A (Imp s t) -> ndc A s -> ndc A t
  | ndcAI A s t:   ndc A s -> ndc A t -> ndc A (And s t)
  | ndcAE A s t u: ndc A (And s t) -> ndc (s::t::A) u -> ndc A u
  | ndcOIL A s t:  ndc A s -> ndc A (Or s t)
  | ndcOIR A s t:  ndc A t -> ndc A (Or s t)
  | ndcOE A s t u: ndc A (Or s t) -> ndc (s::A) u -> ndc (t::A) u -> ndc A u.

Hint Constructors ndc.

Lemma ndc_refl: Reflexivity ndc.
Proof. intros A s. auto. Qed.

Lemma ndc_mono: Monotonicity ndc.
Proof.
  intros A A' s C D. revert C. revert A'. induction D; intros A' C; auto.
  - apply ndcIE with (s:=s); auto.
  - apply ndcAE with (s:=s) (t:=t); try apply IHD2; auto.
  - apply ndcOE with (s:=s) (t:=t); auto.
Qed.

Lemma ndcW1 A s t: ndc A s -> ndc (t::A) s.
Proof. apply ndc_mono; auto. Qed.

Lemma ndcW A B s: ndc A s -> A <<= B -> ndc B s.
Proof. intros C D. apply ndc_mono with (A:=A); auto. Qed.

Lemma ndcE A s: ndc A Fal -> ndc A s.
Proof. intros E. apply ndcC, ndcW1, E. Qed.

Lemma ndc_cut: Cut ndc.
Proof. intros A s t C D. apply ndcIE with (s:=s); auto. Qed.

(* Lemma ndc_cons: Consistency nd. --- based on bent/genc *)

Lemma ndc_imp: CharImp ndc.
Proof. split; auto.
  intros C. apply ndcIE with (s:=s); auto.
  apply (ndcW C); auto.
Qed.

Lemma ndc_fal: CharFal ndc.
Proof. split; auto.
  intros C s. apply ndcC. apply (ndcW C); auto.
Qed.

Lemma ndc_and: CharAnd ndc.
Proof. split; auto. intros C u. apply ndcAE. auto. Qed.

Lemma ndc_or: CharOr ndc.
Proof. split; auto. intros C u. apply ndcOE. auto. Qed.

Lemma ndc_subst: Substitution ndc.
Proof.
  intros A s sigma C. induction C; simpl; auto.
  - apply ndcA, in_map_iff. exists s. auto.
  - apply ndcIE with (s:=subst sigma s); auto.
  - apply ndcAE with (s:=subst sigma s) (t:=subst sigma t); auto.
  - apply ndcOE with (s:=subst sigma s) (t:=subst sigma t); auto.
Qed.

Lemma nd_ndc A s: nd A s -> ndc A s.
Proof.
  intros C. induction C; auto.
  - apply ndcC. apply (ndcW IHC); auto.
  - apply ndcIE with (s:=s); auto.
  - apply ndcAE with (s:=s) (t:=t); auto.
  - apply ndcOE with (s:=s) (t:=t); auto.
Qed.

Lemma ndDN A s: nd A s -> nd A (Not (Not s)).
Proof.
  intros C. apply ndII. apply ndIE with (s:=s). auto. apply (ndW C); auto.
Qed.

Lemma Glivenko A s: ndc A s -> nd A (Not (Not s)).
Proof.
  intros C. induction C.
  - apply ndDN, ndA. auto.
  - apply ndII. apply (ndIE IHC). apply ndII. auto.
  - apply ndII. apply ndIE with (s:=Imp s t). auto.
    apply ndII. apply nd_cut with (s:=Not (Not t)).
    + apply (ndW IHC); auto.
    + apply ndE. apply ndIE with (s:=Not t). auto.
      apply ndII. apply ndIE with (s:=Imp s t); auto.
  - apply ndII, ndIE with (s:=Not (Imp s t)).
    + apply (ndW IHC1); auto.
    + apply ndII, ndIE with (s:=Not s).
      * apply (ndW IHC2); auto.
      * apply ndII. apply ndIE with (s:=t); auto.
        apply ndIE with (s:=s); auto.
  - apply ndII, ndIE with (s:=Not s).
    + apply (ndW IHC1); auto.
    + apply ndII, ndIE with (s:=Not t).
      * apply (ndW IHC2); auto.
      * apply ndII, ndIE with (s:=And s t); auto.
  - apply ndII, ndIE with (s:=Not (And s t)).
    + apply (ndW IHC1); auto.
    + apply ndII, ndIE with (s:=Not u); auto.
      apply ndAE with (s:=s) (t:=t). auto.
      apply (ndW IHC2); auto 7.
  - apply ndII, ndIE with (s:=Not s).
    + apply (ndW IHC); auto.
    + apply ndII, ndIE with (s:=Or s t); auto.
  - apply ndII, ndIE with (s:=Not t).
    + apply (ndW IHC); auto.
    + apply ndII, ndIE with (s:=Or s t); auto.
  - apply ndII, ndIE with (s:=Not (Or s t)).
    + apply (ndW IHC1); auto.
    + apply ndII, ndIE with (s:=Not u); auto.
      apply ndOE with (s:=s) (t:=t). auto.
      * apply (ndW IHC2); auto.
      * apply (ndW IHC3); auto.
Qed.

Corollary Glivenko_refute A: nd A Fal <-> ndc A Fal.
Proof. split.
  - apply nd_ndc.
  - intros E. apply Glivenko in E. apply (ndIE E). apply ndII. auto.
Qed.

Lemma ndc_refute A s: ndc A s <-> ndc (Not s :: A) Fal.
Proof. split ; intros E.
  - apply ndcIE with (s:=s). auto.
    apply (ndcW E); auto.
  - apply ndcC, E.
Qed.

Corollary nd_embeds_ndc A s: ndc A s <-> nd (Not s :: A) Fal.
Proof. setoid_rewrite ndc_refute. symmetry. apply Glivenko_refute. Qed.


(** * Boolean Entailment *)

(** Assignments and Satisfaction *)

Definition assn := var -> bool.

Fixpoint satis (f : assn) (s : form) : Prop :=
  match s with
    | Var x => if f x then True else False
    | Fal => False
    | Imp s1 s2 => satis f s1 -> satis f s2
    | And s1 s2 => satis f s1 /\ satis f s2
    | Or s1 s2 => satis f s1 \/ satis f s2
  end.

Instance satis_dec : forall f s, dec (satis f s).
Proof.
  intros f s. induction s; simpl; auto.
  destruct (f v); auto.
Qed.

Fixpoint eval alpha s : bool :=
  match s with
    | Var x => alpha x
    | Fal => false
    | Imp s t => if eval alpha s then eval alpha t else true
    | And s t => if eval alpha s then eval alpha t else false
    | Or s t => if eval alpha s then true else eval alpha t
  end.

Lemma satis_eval alpha s :
  satis alpha s <-> eval alpha s = true.
Proof.
  induction s; simpl; try (destruct (eval alpha s1); intuition).
  - destruct (alpha v); intuition.
  - intuition.
Qed.

Lemma satis_pos_imp alpha s t :
  satis alpha (Imp s t) <-> ~ satis alpha s \/ satis alpha t.
Proof. simpl. decide (satis alpha s) ; tauto. Qed.

Lemma satis_neg_imp alpha s t :
  ~ satis alpha (Imp s t) <-> satis alpha s /\ ~ satis alpha t.
Proof. simpl. decide (satis alpha s) ; tauto. Qed.

Lemma satis_pos_and alpha s t :
  satis alpha (And s t) <-> satis alpha s /\ satis alpha t.
Proof. simpl. tauto. Qed.

Lemma satis_neg_and alpha s t :
  ~ satis alpha (And s t) <-> ~ satis alpha s \/ ~ satis alpha t.
Proof. simpl. decide (satis alpha s) ; tauto. Qed.

Lemma satis_pos_or alpha s t :
  satis alpha (Or s t) <-> satis alpha s \/ satis alpha t.
Proof. simpl. tauto. Qed.

Lemma satis_neg_or alpha s t :
  ~ satis alpha (Or s t) <-> ~ satis alpha s /\ ~ satis alpha t.
Proof. simpl. tauto. Qed.

(** Validity and boolean entailment *)

Definition valid s : Prop :=
  forall alpha, satis alpha s.

Lemma valid_unsat s :
  valid s <-> ~ exists alpha, satis alpha (Not s).
Proof.
  rewrite DM_exists. split.
  - intros F alpha G. apply G, F.
  - intros F alpha. apply dec_DN; auto. apply F.
Qed.

Definition bent A s : Prop :=
  forall alpha, (forall t, t el A -> satis alpha t) -> satis alpha s.

Lemma ndc_bent A s: ndc A s -> bent A s.
Proof.
  intros E alpha F.
  induction E; simpl in *; auto.
  - decide (satis alpha s) as [G|G]. auto. exfalso. apply IHE.
    intros t [H|H]; subst; auto.
  - intros G. apply IHE. intros u [H|H]; subst; auto.
  - apply IHE2. intros w [H|[H|H]]; subst; auto; tauto.
  - destruct (IHE1 F) as [G|G].
    + apply IHE2. intros w [H|H]; subst; auto.
    + apply IHE3. intros w [H|H]; subst; auto.
Qed.

Lemma ndc_cons: Consistency ndc.
Proof.
  exists Fal. intros C. apply ndc_bent in C.
  apply (C (fun _ => true)). firstorder.
Qed.

Lemma ndc_EntailRelation: EntailRelAllProps ndc.
Proof. split.
  apply ndc_refl. split.
  apply ndc_mono. split.
  apply ndc_cut. split.
  apply ndc_cons. split.
  apply ndc_imp. split.
  apply ndc_fal. split.
  apply ndc_and. split.
  apply ndc_or.
  apply ndc_subst.
Qed.

Implicit Types x y z : var.
Implicit Types s t : form.
Implicit Types alpha beta : assn.
Implicit Types A B : context.
Implicit Types m n k : nat.
Implicit Types S T : sform.
Implicit Types C D : clause.

Fixpoint satis' alpha C : Prop :=
  match C with
    | nil => True
    | T::C' => satis alpha (uns T) /\ satis' alpha C'
  end.

Definition sat C := exists alpha, satis' alpha C.

Lemma satis_iff alpha C :
  satis' alpha C <-> forall S, S el C -> satis alpha (uns S).
Proof.
  induction C; simpl; intuition; subst; auto.
Qed.

Lemma satis_in alpha C S :
  satis' alpha C -> S el C -> satis alpha (uns S).
Proof. 
  intros E. apply satis_iff, E.
Qed.

Lemma satis_weak alpha C C' :
  C <<= C' -> satis' alpha C' -> satis' alpha C.
Proof. 
  rewrite !satis_iff. auto.
Qed.

Lemma sat_weak C C' :
  C <<= C' -> sat C' -> sat C.
Proof. 
  intros E [alpha F]. exists alpha. eauto using satis_weak.
Qed.

(** Entailment of clauses *)

Definition cent C D : Prop :=
  forall alpha, satis' alpha C -> satis' alpha D.

Lemma cent_weak C D D' :
  D <<= D' -> cent C D' -> cent C D.
Proof.
  intros F G alpha K. apply (satis_weak F), G, K.
Qed.

Lemma cent_sat C D :
  cent C D -> sat C -> sat D.
Proof.
  intros F [alpha G]. exists alpha. apply F, G.
Qed.

(** Solved Clauses *)

Definition svar S : Prop :=
  match S with
    | +Var _ | -Var _ => True
    | _ => False
  end.

Definition solved C : Prop :=
  (forall S, S el C -> svar S) /\ forall x, +Var x el C -> ~ -Var x el C.

Definition phi C x := if decision (+Var x el C) then true else false.

Lemma solved_phi C :
  solved C -> satis' (phi C) C.
Proof.
  intros [F G]. apply satis_iff.
  intros S K. assert (L := F S K).
  destruct S as [[x| | | |]|[x| | | |]];
    inv L; simpl; unfold phi;
    decide (+Var x el C) as [L|L]; auto.
  apply G in L. auto.
Qed.

Lemma solved_sat C: solved C -> sat C.
Proof.
  intros F. exists (phi C). apply solved_phi, F.
Qed.

Lemma solved_sat' C: solved C -> { alpha | satis' alpha C }.
Proof.
  intros F. exists (phi C). apply solved_phi, F.
Qed.

Lemma solved_nil :
  solved nil.
Proof.
  firstorder.
Qed.

Lemma solved_pos_var x C :
  solved C -> ~ -Var x el C -> solved (+Var x :: C).
Proof.
  intros F G. split.
  - intros S [K|K].
    + subst S. simpl; auto.
    + apply F, K.
  - intros y [K|K] [L|L]; try inv K; try inv L; auto.
    eapply F; eauto.
Qed.

Lemma solved_neg_var x C :
  solved C -> ~ +Var x el C -> solved (-Var x :: C).
Proof.
  intros F G. split.
  - intros S [K|K].
    + subst S. simpl; auto.
    + apply F, K.
  - intros y [K|K] [L|L]; try inv K; try inv L; auto.
    eapply F; eauto.
Qed.

Lemma cent_solved_sat C D :
  cent C D -> solved C -> sat D.
Proof.
  intros G F. apply (cent_sat G), solved_sat, F.
Qed.

(** * Refutation predicates *)

Record ref_pred (rho : clause -> Prop) :=
  { ref_Fal:     forall C, +Fal el C -> rho C;
    ref_weak:    forall C C', C <<= C' -> rho C -> rho C';
    ref_clash:   forall x C, +Var x el C -> -Var x el C -> rho C;
    ref_pos_imp: forall s t C, rho (-s::C) -> rho (+t::C) -> rho (+Imp s t::C);
    ref_neg_imp: forall s t C, rho (+s::-t::C) -> rho (-Imp s t::C);
    ref_pos_and: forall s t C, rho (+s::+t::C) -> rho (+And s t::C);
    ref_neg_and: forall s t C, rho (-s::C) -> rho (-t::C) -> rho (-And s t::C);
    ref_pos_or:  forall s t C, rho (+s::C) -> rho (+t::C) -> rho (+Or s t::C);
    ref_neg_or:  forall s t C, rho (-s::-t::C) -> rho (-Or s t::C);
    ref_sound:   forall C, rho C -> ~ sat C }.

Lemma ref_unsat :
  ref_pred (fun C => ~ sat C).
Proof.
  split.
  - intros C E [alpha F]. apply (satis_in F E).
  - intros C C' A E F. apply E. revert A F. apply sat_weak.
  - intros s C A B [alpha E]. apply (satis_in E B), (satis_in E A).
  - intros s t C A B [alpha [E F]]. apply satis_pos_imp in E as [E|E].
    + apply A. exists alpha. simpl. auto.
    + apply B. exists alpha. simpl. auto.
  - intros s t C A [alpha [E F]]. apply A. exists alpha. simpl in *.
    apply satis_neg_imp in E as [E G]. auto.
  - intros s t C A [alpha [[E1 E2] F]].
    apply A. exists alpha. simpl. auto.
  - intros s t C A B [alpha [E F]]. apply (satis_neg_and alpha s t) in E as [E|E].
    + apply A. exists alpha. simpl. auto.
    + apply B. exists alpha. simpl. auto.
  - intros s t C A B [alpha [[E|E] F]].
    + apply A. exists alpha. simpl. auto.
    + apply B. exists alpha. simpl. auto.
  - intros s t C A [alpha [E F]]. apply (satis_neg_or alpha s t) in E as [E1 E2].
    apply A. exists alpha. simpl. auto.
  - auto.
Qed.

Lemma ref_ndc :
  ref_pred (fun C => ndc (map uns C) Fal).
Proof.
  split; simpl.
  - intros C D. apply ndcA. exact (in_map uns _ _ D).
  - intros C C' D. apply ndc_mono. apply incl_map, D.
  - intros x C D E. apply ndcIE with (s:=Var x) ; apply ndcA.
    + exact (in_map uns _ _ E).
    + exact (in_map uns _ _ D).
  - intros s t C D E. apply ndcIE with (s:= Not s).
    + apply ndcW with (A:= map uns C); auto.
    + apply ndcII. apply ndcIE with (s:=t).
      * apply ndcW with (A:=map uns C); auto.
      * apply ndcIE with (s:=s) ; auto.
  - intros s t C D. apply ndcIE with (s:= Imp s t). auto.
    apply ndcII, ndcE. apply ndcIE with (s:= Not t).
    + apply ndcIE with (s:= s); auto.
      apply ndcW with (A:=map uns C); auto. apply ndcII, ndcII.
      revert D. apply ndc_mono. auto.
    + apply ndcII. apply ndcIE with (s:= Imp s t); auto.
  - intros s t C D. apply ndcAE with (s:=s) (t:=t). auto.
    apply (ndcW D); auto 6.
  - intros s t C D E. apply ndcIE with (s:=And s t).
    auto. apply ndcAI; apply ndcC.
    + apply (ndcW D); auto.
    + apply (ndcW E); auto.
  - intros s t C D E. apply ndcOE with (s:=s) (t:=t); auto.
    + apply (ndcW D); auto.
    + apply (ndcW E); auto.
  - intros s t C D. apply ndcIE with (s:=s).
    + apply ndcII. apply ndcIE with (s:=Or s t); auto.
    + apply ndcC. apply ndcIE with (s:=t).
      * apply ndcII. apply ndcIE with (s:=Or s t); auto.
      * apply ndcC. apply (ndcW D); auto 6.
  - intros C F [alpha G]. apply ndc_bent in F. apply (F alpha).
    intros t K. apply in_map_iff in K as [[u|u] [K1 K2]]; 
    subst t; revert K2; apply satis_in, G.
Qed.

(** * Decision trees *)

Inductive rec C : clause -> Type :=
  | recNil :       rec C nil
  | recPF D :      rec C (+Fal :: D)
  | recNF D :      rec C D -> rec C (-Fal ::D)
  | recPV D x :    -Var x el C -> rec C (+Var x :: D)
  | recPV' D x :   ~ -Var x el C -> rec (+Var x :: C) D -> rec C (+Var x :: D)
  | recNV D x :    +Var x el C -> rec C (-Var x :: D)
  | recNV' D x :   ~ +Var x el C -> rec (-Var x :: C) D -> rec C (-Var x :: D)
  | recPI D s t :  rec C (-s :: D) -> rec C (+t :: D) -> rec C (+Imp s t :: D)
  | recNI D s t :  rec C (+s :: -t :: D) -> rec C (-Imp s t :: D)
  | recPA D s t : rec C (+s :: +t :: D) -> rec C (+And s t :: D)
  | recNA D s t :  rec C (-s :: D) -> rec C (-t :: D) -> rec C (-And s t :: D)
  | recPO D s t :  rec C (+s :: D) -> rec C (+t :: D) -> rec C (+Or s t :: D)
  | recNO D s t :  rec C (-s :: -t :: D) -> rec C (-Or s t :: D).


Definition provider C D :  rec C D.
Proof.
  revert D C. refine (size_recursion size _).
  intros [|[[x| |s t|s t|s t]|[x| |s t|s t|s t]] D] IH C;
  try (constructor ; apply IH ; simpl ; omega).
  - decide (-Var x el C) as [E|E].
    + apply (recPV _ E).
    + apply (recPV' E). apply IH. simpl ; omega.
  - decide (+Var x el C) as [E|E].
    + apply (recNV _ E).
    + apply (recNV' E). apply IH. simpl ; omega.
Defined.

Definition DNF Cs: list clause :=
  (fix F {C D: clause} (r: rec C D): list clause :=
    match r with
    | recNil => [C]
    | recPF _ => []
    | recNF _ R => F R
    | recPV _ _ _ => []
    | recPV' _ _ _ R => F R
    | recNV _ _ _ => []
    | recNV' _ _ _ R => F R
    | recPI _ _ _ R1 R2 => (F R1) ++ (F R2)
    | recNI _ _ _ R => F R
    | recPA _ _ _ R => F R
    | recNA _ _ _ R1 R2 => (F R1) ++ (F R2)
    | recPO _ _ _ R1 R2 => (F R1) ++ (F R2)
    | recNO _ _ _ R => F R
    end
  ) [] Cs (provider [] Cs).

(* Compute DNF [+ Imp (Var 0) (Imp (Var 1) (Var 0))]. *)

(** * Refutation lemma *)

Section RL.
  Variable rho : clause -> Prop.
  Variable Rho : ref_pred rho.

  Definition decider C D :
    rec C D -> solved C ->
    {E | solved E /\ cent E (D ++ C)} + {rho (D ++ C)}.
  Proof.
    intros F G.
    induction F as [C|C D|C D _ IH|C D x F|C D x F _ IH|C D x F|C D x F _ IH|
                    C D s t _ IHs _ IHt|C D s t _ IH|
                    C D s t _ IH|C D s t _ IHs _ IHt|
                    C D s t _ IHs _ IHt| C D s t _ IH] ; simpl.
    - left. exists C. firstorder.
    - right. apply ref_Fal; auto.
    - destruct (IH G) as [[E [F1 F2]]|F].
      + left. exists E. split. exact F1.
        revert F2. unfold cent. simpl. auto.
      + right. revert F. apply ref_weak; auto.
    - right. apply ref_clash with (x:=x); auto.
    - destruct IH as [[E [K1 K2]]|K].
      + apply solved_pos_var; auto.
      + left. exists E. split. exact K1.
        revert K2. apply cent_weak; auto.
      + right. revert K. apply ref_weak; auto.
    - right. apply ref_clash with (x:=x); auto.
    - destruct IH as [[E [K1 K2]]|K].
      + apply solved_neg_var; auto.
      + left. exists E. split. exact K1.
        revert K2. apply cent_weak; auto.
      + right. revert K. apply ref_weak; auto.
    - destruct (IHs G) as [[E [K1 K2]]|K].
      + left. exists E. split. exact K1.
        intros alpha L. apply K2 in L. simpl in *. tauto.
      + destruct (IHt G) as [[E [L1 L2]]|L].
        * left. exists E. split. exact L1.
          intros alpha L. apply L2 in L. simpl in *. tauto.
        * right. apply ref_pos_imp; auto.
    - destruct (IH G) as [[E [K1 K2]]|K].
      + left. exists E. split. exact K1.
        intros alpha L. apply K2 in L. simpl in *. tauto.
      + right. apply ref_neg_imp; auto.
    - destruct (IH G) as [[E [K1 K2]]|K].
      + left. exists E. split. exact K1.
        intros alpha L. apply K2 in L. simpl in *. tauto.
      + right. apply ref_pos_and; auto.
    - destruct (IHs G) as [[E [K1 K2]]|K].
      + left. exists E. split. exact K1.
        intros alpha L. apply K2 in L. simpl in *. tauto.
      + destruct (IHt G) as [[E [L1 L2]]|L].
        * left. exists E. split. exact L1.
          intros alpha L. apply L2 in L. simpl in *. tauto.
        * right. apply ref_neg_and; auto.
    - destruct (IHs G) as [[E [K1 K2]]|K].
      + left. exists E. split. exact K1.
        intros alpha L. apply K2 in L. simpl in *. tauto.
      + destruct (IHt G) as [[E [L1 L2]]|L].
        * left. exists E. split. exact L1.
          intros alpha L. apply L2 in L. simpl in *. tauto.
        * right. apply ref_pos_or; auto.
    - destruct (IH G) as [[E [K1 K2]]|K].
      + left. exists E. split. exact K1.
        intros alpha L. apply K2 in L. simpl in *. tauto.
      + right. apply ref_neg_or; auto.
  Defined.

  Lemma solved_ref C :
    {D| solved D /\ cent D C} + {rho C}.
  Proof.
    destruct (decider (provider nil C)) as [[E [F1 F2]]|F].
    - apply solved_nil.
    - left. exists E. revert F2. simpl_list. auto.
    - right. revert F. simpl_list. auto.
  Defined.

  Lemma refutation C :
    {sat C} + {rho C}.
  Proof.
    destruct (solved_ref C) as [[D [F1 F2]]|F];
    eauto using cent_solved_sat.
  Defined.

  Lemma ref_iff_unsat C :
    rho C <-> ~ sat C.
  Proof.
    generalize (ref_sound Rho (C:=C)), (refutation C); tauto.
  Qed.

  Lemma ref_dec C :
    dec (rho C).
  Proof.
    unfold dec.
    generalize (refutation C), (ref_iff_unsat C); tauto.
  Defined.

End RL.

(** * Decidability of classical ND entailment *)

Lemma satis_pos f A :
  satis' f (map Pos A) <-> forall s, s el A -> satis f s.
Proof.
  induction A; simpl; firstorder; subst; auto.
Qed.

Theorem ndc_dec_bent A s :
  dec (ndc A s).
Proof.
  generalize (ref_dec ref_ndc (-s :: map Pos A)).
  simpl. rewrite uns_pos. apply dec_prop_iff.
  symmetry. apply ndc_refute.
Defined.

(* Recursive Extraction ndc_dec_bent. *)

(** * Agreement of classical ND entailment with boolean entailment *)

Lemma bent_iff_unsat A s :
  bent A s <-> ~ sat (-s :: map Pos A).
Proof.
  split.
  - intros D [f [E F]]. apply E, (D f). apply satis_pos, F.
  - intros D f E. decide (satis f s) as [F|F]. exact F. exfalso.
    apply D. exists f. split. exact F. apply satis_pos, E.
Qed.

Theorem ndc_iff_sem A s:
  ndc A s <-> bent A s.
Proof.
  rewrite bent_iff_unsat.
  rewrite ndc_refute.
  generalize (ref_iff_unsat ref_ndc (-s :: map Pos A)). 
  simpl. rewrite uns_pos.
  auto.
Qed.

(** * Classical Gentzen system *)

(** GK3c system *)
Inductive gk3c: context -> context -> Prop :=
  | gk3cC A B x:     Var x el A -> Var x el B -> gk3c A B
  | gk3cF A B:       Fal el A -> gk3c A B
  | gk3cIL A B s t:  Imp s t el A -> gk3c A (s::B) -> gk3c (t::A) B -> gk3c A B
  | gk3cIR A B s t:  Imp s t el B -> gk3c (s::A) (t::B) -> gk3c A B
  | gk3cAL A B s t:  And s t el A -> gk3c (s::t::A) B -> gk3c A B
  | gk3cAR A B s t:  And s t el B -> gk3c A (s::B) -> gk3c A (t::B) -> gk3c A B
  | gk3cOL A B s t:  Or s t el A -> gk3c (s::A) B -> gk3c (t::A) B -> gk3c A B
  | gk3cOR A B s t:  Or s t el B -> gk3c A (s::t::B) -> gk3c A B.

Hint Constructors gk3c.

Lemma gk3cA A B s: s el A -> s el B -> gk3c A B.
Proof.
  revert A B. induction s as [x| |s IHs t IHt|s IHs t IHt|s IHs t IHt];
  intros A B E F.
  - eauto.
  - eauto.
  - apply (gk3cIL E); apply gk3cIR with (s:=s) (t:=t); auto.
  - apply (gk3cAL E); apply gk3cAR with (s:=s) (t:=t); auto.
  - apply (gk3cOL E); apply gk3cOR with (s:=s) (t:=t); auto.
Qed.

Hint Resolve gk3cA.

Lemma gk3cLW A B: gk3c A B -> forall A', A <<= A' -> gk3c A' B.
Proof. intros E. induction E; eauto 9. Qed.

Lemma gk3cRW A B: gk3c A B -> forall B', B <<= B' -> gk3c A B'.
Proof. intros E. induction E; eauto 9. Qed.

Lemma gk3cW A B A' B': gk3c A B -> A <<= A' -> B <<= B' -> gk3c A' B'.
Proof. intros E F G. apply gk3cLW with (A:=A); auto. apply (gk3cRW E), G. Qed.


(** Contraction Rules admissible *)
Lemma gk3cLC A B s: gk3c (s::s::A) B -> gk3c (s::A) B.
Proof.
  intros E. remember (s::s::A) as Ax.
  destruct E; subst; eauto using gk3cLW.
Qed.

Lemma gk3cRC A B s: gk3c A (s::s::B) -> gk3c A (s::B).
Proof.
  intros E. remember (s::s::B) as Bx.
  destruct E; subst; eauto using gk3cRW.
Qed.

(*
  Troelstra & Schwichtenberg proved weakening, inversion lemmas,
  and contraction with maximum depth n of derivation trees.
  Here we have not formalized that.
*)

(* Contradiction Rules *)
Lemma gk3cFL A B s: gk3c A (s::B) -> gk3c (Not s::A) B.
Proof. intros E. apply gk3cIL with (s:=s) (t:=Fal); eauto using gk3cLW. Qed.

Lemma gk3cFR A B s: gk3c (s::A) B -> gk3c A (Not s::B).
Proof. intros E. apply gk3cIR with (s:=s) (t:=Fal); eauto using gk3cRW. Qed.

Lemma gk3cDNR A B s: gk3c A (s::B) -> gk3c A (Not (Not s)::B).
Proof.
  intros E. apply gk3cFR. apply gk3cIL with (s:=s) (t:=Fal); eauto using gk3cLW.
Qed.

Lemma gk3cDNL A B s: gk3c (s::A) B -> gk3c (Not (Not s)::A) B.
Proof. intros E. apply gk3cFL, gk3cFR, E. Qed.

Lemma gk3c_emptyR A: ~ gk3c A [Fal] -> ~ gk3c A [].
Proof.
  intros E F. remember [] as B. induction F; subst; eauto using gk3cRW.
Qed.

Lemma gk3cTF' A B: A === [Imp Fal Fal] -> B === [Fal] -> ~ gk3c A B.
Proof.
  intros C D E. induction E;
  try (rewrite D in H; destruct H; auto; congruence); rewrite C in H;
  try (destruct H; auto; congruence).
  destruct H as [H|H]; auto. inv H. apply IHE1. auto. rewrite D. auto.
Qed.

Lemma gk3cTF: ~gk3c [Imp Fal Fal] [Fal].
Proof.
  remember [Imp Fal Fal] as A. remember [Fal] as B.
  apply gk3cTF'.
  - rewrite HeqA. auto.
  - rewrite HeqB. auto.
Qed.

Lemma gk3c_NF': ~ gk3c [] [Fal].
Proof. intros E. apply gk3cTF. eauto using gk3cLW. Qed.

Lemma gk3c_NF1 B: B === [Fal] -> ~ gk3c [] B.
Proof.
  intros D E. remember [] as A.
  induction E; subst; auto; try rewrite D in H; try inv H; try congruence; auto.
Qed.

Lemma gk3c_NF: ~ gk3c [] [Fal].
Proof. intros E. apply gk3c_NF1 with (B:=[Fal]); auto. Qed.

Lemma gk3c_con: ~ gk3c [] [].
Proof. intros E. apply gk3cTF. eauto using gk3cW. Qed.


Lemma gk3c_NV' B x: B === [Var x] -> ~ gk3c [] B.
Proof.
  intros D E. remember [] as A.
  induction E; subst; auto; try rewrite D in H; try inv H; try congruence; auto.
Qed.

Lemma gk3c_NV x: ~ gk3c [] [Var x].
Proof. intros E. apply gk3c_NV' with (B:=[Var x]) (x:=x); auto. Qed.


(** * Equivalence between GK3c and Nc *)

(* From GK3c to Nc *)

Fixpoint neg (B: context): context :=
  match B with
   | [] => []
   | s :: B' => Not s :: neg B'
  end.

Lemma neg_in B s: s el B -> Not s el neg B.
Proof.
  induction B as [|t B]; intros C.
  - exfalso. auto.
  - simpl. destruct C; subst; auto.
Qed.

Lemma neg_incl B B': B <<= B' -> neg B <<= neg B'.
Proof.
  revert B'. induction B as [|t B]; intros B'; simpl; auto.
  intros D w [E|E]; subst.
  - auto using neg_in.
  - apply IHB; auto. apply incl_lcons in D as [_ D]. exact D.
Qed.

Lemma gk3c_ndcG A B:
  gk3c A B -> ndc (A ++ neg B) Fal.
Proof.
  intros E. induction E as [A B x E F|A B E|A B s t E _ IHF _ IHG|
                            A B s t E _ IH|A B s t E _ IH|
                            A B s t E _ IHF _ IHG|A B s t E _ IHF _ IHG|
                            A B s t E _ IH]; auto.
  - apply ndcIE with (s:=Var x) (t:=Fal); apply ndcA; auto.
    apply in_app_iff. right. apply neg_in, F.
  - apply ndc_cut with (s:=t); auto.
    apply ndcIE with (s:=s) (t:=t); auto.
    apply ndcC. apply (ndcW IHF); unfold Not; simpl; auto.
  - apply ndcIE with (s:=Imp s t) (t:=Fal).
    + apply ndcA. apply in_app_iff. right. apply neg_in. auto.
    + apply ndcII. apply ndcC. apply (ndcW IH); simpl; auto.
      rewrite equi_swap. auto 7.
  - apply ndcAE with (s:=s) (t:=t); auto.
  - apply ndcIE with (s:=And s t) (t:=Fal).
    + apply ndcA. apply in_app_iff. right. apply neg_in. auto.
    + apply ndcAI; apply ndcC.
      * apply (ndcW IHF); simpl; auto.
      * apply (ndcW IHG); simpl; auto.
  - apply ndcOE with (s:=s) (t:=t); auto.
  - apply ndc_cut with (s:=Not t).
    + apply ndcII. apply ndcIE with (s:=Or s t) (t:=Fal); auto.
      apply ndcA. right. apply in_app_iff. right. apply neg_in. auto.
    + apply ndc_cut with (s:=Not s).
      * apply ndcII. apply ndcIE with (s:=Or s t) (t:=Fal); auto.
        apply ndcA. right. right. apply in_app_iff. right. apply neg_in. auto.
      * apply (ndcW IH); simpl; auto. intros w G.
        apply in_app_iff in G as [G|[G|[G|G]]]; subst; auto.
Qed.

Lemma gk3c_ndc A s:
  gk3c A [s] -> ndc A s.
Proof.
  intros E. apply gk3c_ndcG in E.
  apply ndcC. apply (ndcW E); simpl; auto.
Qed.

(** Assumming Cut *)
Definition gk3cCut: Prop :=
  forall A B A' B' s, gk3c A (s::B) -> gk3c (s::A') B' -> gk3c (A++A') (B++B').

Section gk3c_Cut.
  Implicit Type Cut: gk3cCut.

  (* Explosion Rule or Falsity Characterization *)
  Lemma gk3cCutE Cut A B: gk3c A [Fal] -> gk3c A B.
  Proof.
    intros E. apply gk3cW with (A:=A ++ []) (B:= [] ++ B); auto.
    apply (@Cut A [] [] B Fal); auto.
  Qed.

  Lemma gk3cFL_invCut Cut A B s:
    gk3c (Not s::A) B -> gk3c A (s::B).
  Proof.
    intros E. apply gk3cW with (A:=A ++ A) (B:=(s::B) ++ B); auto.
    apply (@Cut A (s::B) A B (Not s)); auto.
    apply gk3cFR. eauto.
  Qed.

  Lemma gk3cFR_invCut Cut A B s:
    gk3c A (Not s::B) -> gk3c (s::A) B.
  Proof.
    intros E. apply gk3cW with (A:=A ++ (s::A)) (B:=B ++ B); auto.
    apply (@Cut A B (s::A) B (Not s)); auto.
    apply gk3cFL. eauto.
  Qed.

  Lemma gk3cDNR_invCut Cut A B s:
    gk3c A (Not (Not s)::B) -> gk3c A (s::B).
  Proof.
    intros E. apply gk3cW with (A:=A ++ A) (B:=B ++ s :: B); auto.
    apply (@Cut A B A (s::B) (Not (Not s)) E).
    apply gk3cDNL. eauto.
  Qed.

  Lemma gk3cDNL_invCut Cut A B s:
    gk3c (Not (Not s)::A) B -> gk3c (s::A) B.
  Proof.
    intros E.
    apply gk3cW with (A:= (s :: A) ++ A) (B:=B ++ B); auto.
    apply (@Cut (s::A) B A B (Not (Not s))); eauto using gk3cW.
    apply gk3cDNR. eauto.
  Qed.

  Example gk3c_Peirce_Cut Cut s t: gk3c [] [(Imp (Imp (Imp s t) s) s)].
  Proof.
    apply gk3cIR with (s:=Imp (Imp s t) s) (t:=s); auto.
    apply gk3cRW with (B:=[s]); auto.
    apply (gk3cDNR_invCut Cut).
    apply gk3cIL with (s:=Imp s t) (t:=s); auto.
    - apply gk3cIR with (s:=s) (t:=t); auto.
      apply gk3cW with (A:=[s; Imp (Imp s t) s]) (B:=[Not (Not s); t; Imp s t]);
      auto. apply gk3cDNR. eauto.
    - apply gk3cDNR. eauto.
  Qed.

  (* Equivalence with natural deduction, assuming Cut *)
  Lemma ndc_gk3cCut Cut A s:
    ndc A s -> gk3c A [s].
  Proof.
    intros E. induction E; eauto using gk3cW.
    - apply gk3cFR in IHE. apply (gk3cDNR_invCut Cut) in IHE.
      apply gk3cW with (A:=A ++ []) (B:=[s] ++ [s]); auto.
      apply (@Cut A [s] [] [s] Fal); auto.
      apply gk3cW with (A:=A) (B:=[s;Fal]); auto.
    - apply gk3cW with (A:=A ++ A) (B:=[] ++ [t]); auto.
      apply (@Cut A [] A [t] s); auto.
      apply gk3cW with (A:=(s::A) ++ [s]) (B:=[t]); auto.
      apply (@Cut (s::A) [] [s] [t] (Imp s t)); eauto using gk3cW.
    - apply gk3cAR with (s:=s) (t:=t); eauto using gk3cW.
    - apply gk3cW with (A:=A ++ A) (B:=[u]); auto.
      apply (@Cut A [] A [u] (And s t)); auto.
      apply gk3cAL with (s:=s) (t:=t); eauto using gk3cW, incl_shift.
    - apply gk3cW with (A:=A ++ A) (B:=[u]); auto.
      apply (@Cut A [] A [u] (Or s t)); auto.
      apply gk3cOL with (s:=s) (t:=t); eauto using gk3cW.
  Qed.

  Theorem gk3cCut_iff_ndc Cut A s:
    gk3c A [s] <-> ndc A s.
  Proof. split.
    - apply gk3c_ndc.
    - apply ndc_gk3cCut, Cut.
  Qed.

End gk3c_Cut.


(** * Cut elimination in GK3c *)

Proposition gk3c_GCut_Var A B x:
  gk3c A B ->
  forall A' B' : context, gk3c A' B' ->
    gk3c (A ++ rem A' (Var x)) (rem B (Var x) ++ B').
Proof.
  intros C Ay By D.
  induction C as [A B y E F| |A B s t E F IHF G IHG|A B s t E F IH|
                  A B s t E F IH|A B s t E F IHF G IHG|A B s t E F IHF G IHG|
                  A B s t E F IH]; auto.
  - decide (Var y = Var x) as [K|K]; subst.
    + inv K. apply (gk3cW D); auto.
    + apply gk3cA with (s:=Var y); auto.
  - apply gk3cIL with (s:=s) (t:=t); auto. apply (gk3cRW IHF); auto.
  - assert (Imp s t <> Var x) by congruence.
    apply gk3cIR with (s:=s) (t:=t); auto. apply (gk3cRW IH); auto.
  - apply gk3cAL with (s:=s) (t:=t); auto.
  - assert (And s t <> Var x) by congruence.
    apply gk3cAR with (s:=s) (t:=t); auto.
    + apply (gk3cRW IHF); auto.
    + apply (gk3cRW IHG); auto.
  - apply gk3cOL with (s:=s) (t:=t); auto.
  - assert (Or s t <> Var x) by congruence.
    apply gk3cOR with (s:=s) (t:=t); auto. apply (gk3cRW IH); auto 7.
Qed.

Proposition gk3c_GCut_Fal A B:
  gk3c A B ->
  forall A' B' : context, gk3c A' B' ->
    gk3c (A ++ rem A' Fal) (rem B Fal ++ B').
Proof.
  intros C Ay By D.
  induction C as [A B x E F| |A B s t E F IHF G IHG|A B s t E F IH|
                  A B s t E F IH|A B s t E F IHF G IHG|A B s t E F IHF G IHG|
                  A B s t E F IH]; auto.
  - assert (Var x <> Fal) by congruence. apply gk3cA with (s:=Var x); auto.
  - apply gk3cIL with (s:=s) (t:=t); auto. apply (gk3cRW IHF); auto.
  - assert (Imp s t <> Fal) by congruence.
    apply gk3cIR with (s:=s) (t:=t); auto. apply (gk3cRW IH); auto.
  - apply gk3cAL with (s:=s) (t:=t); auto.
  - assert (And s t <> Fal) by congruence. apply gk3cAR with (s:=s) (t:=t); auto.
    + apply (gk3cRW IHF); auto.
    + apply (gk3cRW IHG); auto.
  - apply gk3cOL with (s:=s) (t:=t); auto.
  - assert (Or s t <> Fal) by congruence.
    apply gk3cOR with (s:=s) (t:=t); auto. apply (gk3cRW IH); auto 7.
Qed.

Lemma gk3c_GCut A B A' B' s:
  gk3c A B -> gk3c A' B' -> gk3c (A ++ rem A' s) (rem B s ++ B').
Proof.
  intros C. revert A B C A' B'.
  induction s as [x| |s IHs t IHt|s IHs t IHt|s IHs t IHt]; intros A B C Ay By D.
  - apply gk3c_GCut_Var; auto.
  - apply gk3c_GCut_Fal; auto.
  - induction C as [A B x E F| |A B s1 t1 E F IHF G IHG|A B s1 t1 E F IH|
                    A B s1 t1 E F IH|A B s1 t1 E F IHF G IHG|
                    A B s1 t1 E F IHF G IHG|A B s1 t1 E F IH]; auto.
    + assert (Imp s t <> Var x) by congruence. apply gk3cA with (s:=Var x); auto.
    + apply gk3cIL with (s:=s1) (t:=t1); auto. apply (gk3cRW IHF); auto.
    + decide (Imp s1 t1 = Imp s t) as [H|H].
      * inv H.
      { induction D as [Ay By x E1 F1|Ay By E1|Ay By s1 t1 E1 F1 IHF1 G1 IHG1|
                        Ay By s1 t1 E1 F1 IH1|Ay By s1 t1 E1 F1 IH1|
                        Ay By s1 t1 E1 F1 IHF1 G1 IHG1|
                        Ay By s1 t1 E1 F1 IHF1 G1 IHG1|Ay By s1 t1 E1 F1 IH1].
        - assert (Var x <> Imp s t) by congruence.
          apply gk3cA with (s:=Var x); auto.
        - assert (Fal <> Imp s t) by congruence. apply gk3cF. auto.
        - assert (IHF: gk3c (A ++ rem Ay (Imp s t))
                            (rem B (Imp s t) ++ s1 :: By)).
            { apply IHF1. apply (gk3cRW IH). auto. } clear IHF1.
          assert (IHG: gk3c (A ++ rem (t1 :: Ay) (Imp s t))
                            (rem B (Imp s t) ++ By)).
            { apply IHG1. apply (gk3cLW IH). auto. } clear IHG1.
          decide (Imp s1 t1 = Imp s t) as [H|H].
          + inv H.
            remember (A ++ rem Ay (Imp s t)) as Ax.
            remember (rem B (Imp s t) ++ By) as Bx.
            remember (A ++ rem (t :: Ay) (Imp s t)) as Az.
            remember (rem B (Imp s t) ++ s :: By) as Bz.
            remember ((s :: A) ++ rem Ay (Imp s t)) as Aw.
            remember (rem (t :: B) (Imp s t) ++ By) as Bw.
            assert (K1:= IHs Ax Bz IHF Az Bx IHG).
            assert (K2:= IHs Ax Bz IHF Aw Bw IH).
            assert (K3:= IHt (Ax ++ rem Aw s) (rem Bz s ++ Bw) K2
                             (Ax ++ rem Az s) (rem Bz s ++ Bx) K1).
            apply (gk3cW K3); auto; clear K1 K2 K3.
            * apply incl_app. subst; auto.
              apply rem_app'; auto. rewrite rem_comm.
              assert (rem (rem Az t) s <<= rem Az t).
                { unfold rem. apply filter_incl. } rewrite H.
              subst. apply rem_app'; auto.
              apply incl_appr. rewrite rem_comm. auto.
            * apply incl_app.
              { apply rem_app'.
                - assert (rem (rem Bz s) t <<= rem Bz s).
                  { unfold rem. apply filter_incl. } rewrite H. subst.
                  apply rem_app'; auto.
                - subst. apply rem_app'; auto. rewrite rem_comm. auto. }
              { apply incl_app; auto. subst. apply rem_app'; auto. }
          + apply gk3cIL with (s:=s1) (t:=t1); auto.
            * apply (gk3cW IHF); auto.
            * apply (gk3cW IHG); auto.
        - assert (IH': gk3c (A ++ rem (s1 :: Ay) (Imp s t))
                            (rem B (Imp s t) ++ t1 :: By)).
          { apply IH1. apply (gk3cW IH); auto. } clear IH1.
          apply gk3cIR with (s:=s1) (t:=t1); auto. apply (gk3cW IH'); auto.
        - assert (IH': gk3c (A ++ rem (s1 :: t1 :: Ay) (Imp s t))
                            (rem B (Imp s t) ++ By)).
            { apply IH1. apply (gk3cW IH); auto 6. } clear IH1.
          assert (And s1 t1 <> Imp s t) by congruence.
          apply gk3cAL with (s:=s1) (t:=t1); auto. apply (gk3cLW IH'); auto 7.
        - assert (IHF: gk3c (A ++ rem Ay (Imp s t))
                            (rem B (Imp s t) ++ s1 :: By)).
            { apply IHF1. apply (gk3cRW IH); auto. }
          assert (IHG: gk3c (A ++ rem Ay (Imp s t))
                       (rem B (Imp s t) ++ t1 :: By)).
            { apply IHG1. apply (gk3cRW IH); auto. } clear IHF1 IHG1.
          assert (And s1 t1 <> Imp s t) by congruence.
          apply gk3cAR with (s:=s1) (t:=t1); auto.
          + apply (gk3cRW IHF); auto.
          + apply (gk3cRW IHG); auto.
        - assert (IHF: gk3c (A ++ rem (s1 :: Ay) (Imp s t))
                            (rem B (Imp s t) ++ By)).
            { apply IHF1. apply (gk3cLW IH); auto. }
          assert (IHG: gk3c (A ++ rem (t1 :: Ay) (Imp s t))
                            (rem B (Imp s t) ++ By)).
            { apply IHG1. apply (gk3cLW IH); auto. } clear IHF1 IHG1.
          assert (Or s1 t1 <> Imp s t) by congruence.
          apply gk3cOL with (s:=s1) (t:=t1); auto.
          + apply (gk3cLW IHF); auto.
          + apply (gk3cLW IHG); auto.
        - assert (IH': gk3c (A ++ rem Ay (Imp s t))
                       (rem B (Imp s t) ++ s1 :: t1 :: By)).
            { apply IH1. apply (gk3cRW IH); auto. } clear IH1.
          assert (Or s1 t1 <> Imp s t) by congruence.
          apply gk3cOR with (s:=s1) (t:=t1); auto.
          apply (gk3cRW IH'); auto 7. }
      * apply gk3cIR with (s:=s1) (t:=t1); auto.
        apply (gk3cRW IH); auto.
    + assert (And s1 t1 <> Imp s t) by congruence.
      apply gk3cAL with (s:=s1) (t:=t1); auto.
    + assert (And s1 t1 <> Imp s t) by congruence.
      apply gk3cAR with (s:=s1) (t:=t1); auto.
      * apply (gk3cRW IHF); auto.
      * apply (gk3cRW IHG); auto.
    + assert (Or s1 t1 <> Imp s t) by congruence.
      apply gk3cOL with (s:=s1) (t:=t1); auto.
    + assert (Or s1 t1 <> Imp s t) by congruence.
      apply gk3cOR with (s:=s1) (t:=t1); auto. apply (gk3cRW IH); auto 7.
  - induction C as [A B x E F| |A B s1 t1 E F IHF G IHG|A B s1 t1 E F IH|
                    A B s1 t1 E F IH|A B s1 t1 E F IHF G IHG|
                    A B s1 t1 E F IHF G IHG|A B s1 t1 E F IH]; auto.
    + assert (Var x <> And s t) by congruence. apply gk3cA with (s:=Var x); auto.
    + assert (Imp s1 t1 <> And s t) by congruence.
      apply gk3cIL with (s:=s1) (t:=t1); auto. apply (gk3cRW IHF); auto.
    + assert (Imp s1 t1 <> And s t) by congruence.
      apply gk3cIR with (s:=s1) (t:=t1); auto. apply (gk3cRW IH); auto.
    + apply gk3cAL with (s:=s1) (t:=t1); auto.
    + decide (And s1 t1 = And s t) as [H|H].
      * inv H.
      { induction D as [Ay By x E1 F1|Ay By E1|Ay By s1 t1 E1 F1 IHF1 G1 IHG1|
                        Ay By s1 t1 E1 F1 IH1|Ay By s1 t1 E1 F1 IH1|
                        Ay By s1 t1 E1 F1 IHF1 G1 IHG1|
                        Ay By s1 t1 E1 F1 IHF1 G1 IHG1|Ay By s1 t1 E1 F1 IH1].
        - assert (Var x <> And s t) by congruence.
          apply gk3cA with (s:=Var x); auto.
        - assert (Fal <> And s t) by congruence. apply gk3cF. auto.
        - assert (IHF': gk3c (A ++ rem Ay (And s t)) 
                             (rem B (And s t) ++ s1 :: By)).
            { apply IHF1. apply (gk3cRW IHG); auto. apply (gk3cRW IHF); auto. }
          assert (IHG': gk3c (A ++ rem (t1 :: Ay) (And s t))
                             (rem B (And s t) ++ By)).
            { apply IHG1. apply (gk3cLW IHG); auto. apply (gk3cLW IHF); auto. }
          clear IHF1 IHG1.
          assert (Imp s1 t1 <> And s t) by congruence.
          apply gk3cIL with (s:=s1) (t:=t1); auto.
          + apply (gk3cRW IHF'); auto.
          + apply (gk3cLW IHG'); auto.
        - assert (IH: gk3c (A ++ rem (s1 :: Ay) (And s t))
                           (rem B (And s t) ++ t1 :: By)).
            { apply IH1. apply (gk3cW IHG); auto. apply (gk3cW IHF); auto. }
          clear IH1. assert (Imp s1 t1 <> And s t) by congruence.
          apply gk3cIR with (s:=s1) (t:=t1); auto. apply (gk3cW IH); auto.
        - assert (IH: gk3c (A ++ rem (s1 :: t1 :: Ay) (And s t))
                           (rem B (And s t) ++ By)).
            { apply IH1. apply (gk3cLW IHG); auto 7. apply (gk3cLW IHF); auto 7. }  
          clear IH1. decide (And s1 t1 = And s t) as [H|H].
          + inv H.
            remember (A ++ rem Ay (And s t)) as Ax.
            remember (rem B (And s t) ++ By) as Bx.
            remember (A ++ rem (s :: t :: Ay) (And s t)) as Aw.
            remember (rem (t :: B) (And s t) ++ By) as Bt.
            remember (rem (s :: B) (And s t) ++ By) as Bs.
            assert (H1:= IHt Ax Bt IHG Aw Bx IH).
            assert (H2:= IHs Ax Bs IHF (Ax ++ rem Aw t) (rem Bt t ++ Bx) H1).
            apply (gk3cW H2); auto.
            * apply incl_app; auto. apply rem_app'; auto. subst.
              intros u I. apply in_rem_iff in I as [I2 I1].
              apply in_rem_iff in I2 as [I3 I2].
              apply in_app_iff in I3 as [I3|I3]; auto.
              apply in_rem_iff in I3 as [[I3|[I3|I3]] I4];
              subst; try congruence; auto.
            * apply incl_app.
              { subst. apply rem_app'; auto. rewrite rem_comm. auto. }
              { apply incl_app; auto. subst. apply rem_app'; auto.
                rewrite rem_comm. auto. }
          + apply gk3cAL with (s:=s1) (t:=t1); auto. apply (gk3cLW IH); auto 7.
        - assert (IHF': gk3c (A ++ rem Ay (And s t))
                             (rem B (And s t) ++ s1 :: By)).
            { apply IHF1. apply (gk3cRW IHG); auto. apply (gk3cRW IHF); auto. }
          assert (IHG': gk3c (A ++ rem Ay (And s t))
                             (rem B (And s t) ++ t1 :: By)).
            { apply IHG1. apply (gk3cRW IHG); auto. apply (gk3cRW IHF); auto. }
          clear IHF1 IHG1.
          apply gk3cAR with (s:=s1) (t:=t1); auto.
          + apply (gk3cRW IHF'); auto.
          + apply (gk3cRW IHG'); auto.
        - assert (IHF': gk3c (A ++ rem (s1 :: Ay) (And s t))
                             (rem B (And s t) ++ By)).
            { apply IHF1. apply (gk3cLW IHG); auto. apply (gk3cLW IHF); auto. }
          assert (IHG': gk3c (A ++ rem (t1 :: Ay) (And s t))
                             (rem B (And s t) ++ By)).
            { apply IHG1. apply (gk3cLW IHG); auto. apply (gk3cLW IHF); auto. }
          clear IHF1 IHG1.
          assert (Or s1 t1 <> And s t) by congruence.
          apply gk3cOL with (s:=s1) (t:=t1); auto.
          + apply (gk3cLW IHF'); auto.
          + apply (gk3cLW IHG'); auto.
        - assert (IH: gk3c (A ++ rem Ay (And s t))
                           (rem B (And s t) ++ s1 :: t1 :: By)).
            { apply IH1. apply (gk3cRW IHG); auto. apply (gk3cRW IHF); auto. } 
          clear IH1. apply gk3cOR with (s:=s1) (t:=t1); auto.
          apply (gk3cRW IH); auto 7. }
      * apply gk3cAR with (s:=s1) (t:=t1); auto.
        apply (gk3cRW IHF); auto. apply (gk3cRW IHG); auto.
    + apply gk3cOL with (s:=s1) (t:=t1); auto.
    + assert (Or s1 t1 <> And s t) by congruence.
      apply gk3cOR with (s:=s1) (t:=t1); auto. apply (gk3cRW IH); auto 7.
  - induction C as [A B x E F| |A B s1 t1 E F IHF G IHG|A B s1 t1 E F IH|
                    A B s1 t1 E F IH|A B s1 t1 E F IHF G IHG|
                    A B s1 t1 E F IHF G IHG|A B s1 t1 E F IH]; auto.
    + assert (Var x <> Or s t) by congruence. apply gk3cA with (s:=Var x); auto.
    + assert (Imp s1 t1 <> Or s t) by congruence.
      apply gk3cIL with (s:=s1) (t:=t1); auto. apply (gk3cRW IHF); auto.
    + assert (Imp s1 t1 <> Or s t) by congruence.
      apply gk3cIR with (s:=s1) (t:=t1); auto. apply (gk3cRW IH); auto.
    + apply gk3cAL with (s:=s1) (t:=t1); auto.
    + assert (And s1 t1 <> Or s t) by congruence.
      apply gk3cAR with (s:=s1) (t:=t1); auto.
      * apply (gk3cRW IHF); auto.
      * apply (gk3cRW IHG); auto.
    + apply gk3cOL with (s:=s1) (t:=t1); auto.
    + decide (Or s1 t1 = Or s t) as [H|H].
      * inv H.
      { induction D as [Ay By x E1 F1|Ay By E1|Ay By s1 t1 E1 F1 IHF1 G1 IHG1|
                        Ay By s1 t1 E1 F1 IH1|Ay By s1 t1 E1 F1 IH1|
                        Ay By s1 t1 E1 F1 IHF1 G1 IHG1|
                        Ay By s1 t1 E1 F1 IHF1 G1 IHG1|Ay By s1 t1 E1 F1 IH1].
        - assert (Var x <> Or s t) by congruence.
          apply gk3cA with (s:=Var x); auto.
        - assert (Fal <> Or s t) by congruence. apply gk3cF. auto.
        - assert (IHF': gk3c (A ++ rem Ay (Or s t))
                             (rem B (Or s t) ++ s1 :: By)).
            { apply IHF1. apply (gk3cRW IH); auto. }
          assert (IHG': gk3c (A ++ rem (t1 :: Ay) (Or s t))
                             (rem B (Or s t) ++ By)).
            { apply IHG1. apply (gk3cLW IH); auto. } clear IHF1 IHG1.
          assert (Imp s1 t1 <> Or s t) by congruence.
          apply gk3cIL with (s:=s1) (t:=t1); auto.
          + apply (gk3cRW IHF'); auto.
          + apply (gk3cLW IHG'); auto.
        - assert (IH': gk3c (A ++ rem (s1 :: Ay) (Or s t))
                            (rem B (Or s t) ++ t1 :: By)).
            { apply IH1. apply (gk3cW IH); auto. } clear IH1.
          assert (Imp s1 t1 <> Or s t) by congruence.
          apply gk3cIR with (s:=s1) (t:=t1); auto. apply (gk3cW IH'); auto.
        - assert (IH': gk3c (A ++ rem (s1 :: t1 :: Ay) (Or s t))
                            (rem B (Or s t) ++ By)).
            { apply IH1. apply (gk3cLW IH); auto 7. } clear IH1.
          assert (And s1 t1 <> Or s t) by congruence.
          apply gk3cAL with (s:=s1) (t:=t1); auto. apply (gk3cLW IH'); auto 7.
        - assert (IHF: gk3c (A ++ rem Ay (Or s t)) (rem B (Or s t) ++ s1 :: By)).
          { apply IHF1. apply (gk3cRW IH); auto. }
          assert (IHG: gk3c (A ++ rem Ay (Or s t)) (rem B (Or s t) ++ t1 :: By)).
          { apply IHG1. apply (gk3cRW IH); auto. } clear IHF1 IHG1.
          apply gk3cAR with (s:=s1) (t:=t1); auto.
          + apply (gk3cRW IHF); auto.
          + apply (gk3cRW IHG); auto.
        - assert (IHF: gk3c (A ++ rem (s1 :: Ay) (Or s t)) (rem B (Or s t) ++ By)).
          { apply IHF1. apply (gk3cLW IH); auto. }
          assert (IHG: gk3c (A ++ rem (t1 :: Ay) (Or s t)) (rem B (Or s t) ++ By)).
          { apply IHG1. apply (gk3cLW IH); auto. } clear IHF1 IHG1.
          decide (Or s1 t1 = Or s t) as [H|H].
          + inv H.
            remember (A ++ rem Ay (Or s t)) as Ax.
            remember (rem B (Or s t) ++ By) as Bx.
            remember (rem (s :: t :: B) (Or s t) ++ By) as Bst.
            remember (A ++ rem (t :: Ay) (Or s t)) as At.
            remember (A ++ rem (s :: Ay) (Or s t)) as As.
            assert (H1:= IHt Ax Bst IH At Bx IHG).
            assert (H2:= IHs (Ax ++ rem At t) (rem Bst t ++ Bx) H1 As Bx IHF).
            apply (gk3cW H2); auto.
            * apply incl_app; auto.
              { apply incl_app; auto. subst. apply rem_app'; auto.
                rewrite rem_comm. auto. }
              { subst. apply rem_app'; auto. rewrite rem_comm. auto. }
            * apply incl_app; auto. apply rem_app'; auto.
              subst. intros u I. apply in_rem_iff in I as [I2 I1].
              apply in_rem_iff in I2 as [I3 I2].
              apply in_app_iff in I3 as [I3|I3]; auto.
              apply in_rem_iff in I3 as [[I4|[I4|I4]] I3];
              subst; try congruence; auto.
          + apply gk3cOL with (s:=s1) (t:=t1); auto.
            * apply (gk3cLW IHF); auto.
            * apply (gk3cLW IHG); auto.
        - assert (IH': gk3c (A ++ rem Ay (Or s t))
                       (rem B (Or s t) ++ s1 :: t1 :: By)).
            { apply IH1. apply (gk3cRW IH); auto. } clear IH1.
          apply gk3cOR with (s:=s1) (t:=t1); auto. apply (gk3cRW IH'); auto 7. }
      * apply gk3cOR with (s:=s1) (t:=t1); auto.
        apply (gk3cRW IH); auto 7.
Qed.

Theorem gk3c_Cut: gk3cCut.
Proof.
  intros A B A' B' s C D.
  assert (H:= gk3c_GCut s C D). apply (gk3cW H); auto.
Qed.

(** Lemmas transferred to the cut-free system *)

Lemma gk3cE A B: gk3c A [Fal] -> gk3c A B.
Proof. apply (gk3cCutE gk3c_Cut). Qed.

Lemma gk3cFL_inv A B s:
  gk3c (Not s::A) B -> gk3c A (s::B).
Proof. apply (gk3cFL_invCut gk3c_Cut). Qed.

Lemma gk3cFR_inv A B s:
  gk3c A (Not s::B) -> gk3c (s::A) B.
Proof. apply (gk3cFR_invCut gk3c_Cut). Qed.

Lemma gk3cFL_iff A B s:
   gk3c (Not s::A) B <-> gk3c A (s::B).
Proof. split. apply gk3cFL_inv. apply gk3cFL. Qed.

Lemma gk3cFR_iff A B  s:
  gk3c A (Not s::B) <-> gk3c (s::A) B.
Proof. split. apply gk3cFR_inv. apply gk3cFR. Qed.

Lemma gk3cDNR_inv A B s:
  gk3c A (Not (Not s)::B) -> gk3c A (s::B).
Proof. apply (gk3cDNR_invCut gk3c_Cut). Qed.

Lemma gk3cDNL_inv A B s:
  gk3c (Not (Not s)::A) B -> gk3c (s::A) B.
Proof. apply (gk3cDNL_invCut gk3c_Cut). Qed.

Lemma gk3cDNR_iff A B s:
  gk3c A (Not (Not s)::B) <-> gk3c A (s::B).
Proof. split. apply gk3cDNR_inv. apply gk3cDNR. Qed.

Lemma gk3cDNL_iff A B s:
  gk3c (Not (Not s)::A) B <-> gk3c (s::A) B.
Proof. split. apply gk3cDNL_inv. apply gk3cDNL. Qed.

(* Equivalence with natural deduction, cut-free *)
Lemma ndc_gk3c A s:
  ndc A s -> gk3c A [s].
Proof. apply (ndc_gk3cCut gk3c_Cut). Qed.

Theorem gk3c_iff_ndc A s:
  gk3c A [s] <-> ndc A s.
Proof. apply (gk3cCut_iff_ndc gk3c_Cut). Qed.

(* Implication free formulas are equivalently derivable *)
Definition ImpFreeA A := forall s, s el A -> ImpFree s.

Lemma gk3c_tab_ImpFreeG B:
  ImpFreeA B -> gk3c [] B -> tab [] B.
Proof.
  intros D E. remember [] as A.
  induction E as [A B x E F|A B E|A B s t E F IHF G IHG|
                            A B s t E F IH|A B s t E F IH|
                            A B s t E F IHF G IHG|A B s t E F IHF G IHG|
                            A B s t E F IH];
  subst; try (exfalso; apply E).
  - exfalso. apply (D (Imp s t) E).
  - apply (tabAN E).
    + apply IHF; auto. intros u [H|H]; subst; auto.
      destruct (D (And u t) E). auto.
    + apply IHG; auto. intros u [H|H]; subst; auto.
      destruct (D (And s u) E). auto.
  - apply (tabON E). apply IH; auto. intros u [H|[H|H]]; subst; auto.
    + destruct (D (Or u t) E). auto.
    + destruct (D (Or s u) E). auto.
Qed.

Lemma nd_ndc_ImpFree s:
  ImpFree s -> (ndc [] s <-> nd [] s).
Proof.
  intros D. split.
  - intros E. apply gen_iff_nd. apply gen_iff_tab.
    apply gk3c_iff_ndc in E. revert E. apply gk3c_tab_ImpFreeG.
    intros u [H|H]; subst; auto. exfalso. auto.
  - apply nd_ndc.
Qed.

Section ssL_nd_ndc.
  Lemma gk3c_tab_NegImpFreeG B:
    (~ exists S, match S with | -(Imp _ _) => True | _ => False end /\
                 S el ssL (map Neg B))
    -> gk3c [] B -> tab [] B.
  Proof.
    intros D E. remember [] as A.
    induction E as [A B x E F|A B E|A B s t E F IHF G IHG|
                            A B s t E _ _|A B s t E F IH|
                            A B s t E _ IHF _ IHG|A B s t E F IHF G IHG|
                            A B s t E _ IH];
  subst; try (exfalso; apply E).
  - exfalso. apply D. exists (-(Imp s t)). split; auto.
    apply ssL_in, in_map_iff. exists (Imp s t). split; auto.
  - assert (H: tab [] (s :: B)). {
      apply IHF; auto. intros [[|S] [F1 F2]]; auto.
      destruct S as [| |s1 s2 | |]; auto. simpl in F2.
      pose (K:= ssL_correct (map Neg B)); destruct K as [K1 K2].
      apply in_app_iff in F2 as [F2|F2];
      apply D; exists (-(Imp s1 s2)); split; auto.
      apply ssL'_in_ssL2 with (S:=-s); auto.
      destruct (ssL_correct (map Neg B)) as [K _].
      assert (H3: -And s t el map Neg B).
        { apply in_map_iff. exists (And s t). split; auto. }
      destruct (K (-And s t) (ssL_in H3)) as [K3 _]. exact K3. } clear IHF.
    assert (K: tab [] (t :: B)). {
      apply IHG; auto. intros [[|S] [F1 F2]]; auto.
      destruct S as [| |s1 s2 | |]; auto. simpl in F2.
      pose (K:= ssL_correct (map Neg B)); destruct K as [K1 K2].
      apply in_app_iff in F2 as [F2|F2];
      apply D; exists (-(Imp s1 s2)); split; auto.
      apply ssL'_in_ssL2 with (S:=-t); auto.
      destruct (ssL_correct (map Neg B)) as [K _].
      assert (H3: -And s t el map Neg B).
        { apply in_map_iff. exists (And s t). split; auto. }
      destruct (K (-And s t) (ssL_in H3)) as [_ K3]. exact K3. } clear IHG.
    apply (tabAN E); auto.
  - assert (H: tab [] (s :: t :: B)). {
      apply IH; auto. intros [[|S] [F1 F2]]; auto.
      destruct S as [| |s1 s2 | |]; auto. simpl in F2.
      pose (K:= ssL_correct (map Neg B)); destruct K as [K1 K2].
      assert (H3: -Or s t el map Neg B).
          { apply in_map_iff. exists (Or s t). split; auto. }
      destruct (ssL_correct (map Neg B)) as [K _].
      destruct (K (-Or s t) (ssL_in H3)) as [K3 K4].
      apply in_app_iff in F2 as [F2|F2].
      - apply D. exists (-(Imp s1 s2)). split; auto.
        apply ssL'_in_ssL2 with (S:=-s); auto.
      - apply in_app_iff in F2 as [F2|F2];
        apply D; exists (-(Imp s1 s2)); split; auto.
        apply ssL'_in_ssL2 with (S:=-t); auto. } clear IH.
    apply (tabON E H).
  Qed.

  Lemma gk3c_tab_NegImpFree s:
    (~ exists S, match S with | -(Imp _ _) => True | _ => False end /\
                 S el ssL [-s])
    -> gk3c [] [s] -> tab [] [s].
  Proof. apply (@gk3c_tab_NegImpFreeG [s]). Qed.

  Lemma nd_ndc_NegImpFree s:
    (~ exists S, match S with | -(Imp _ _) => True | _ => False end /\
                 S el ssL [-s])
    -> (ndc [] s <-> nd [] s).
  Proof.
    intros D. split.
    - intros E. apply gen_iff_nd. apply gen_iff_tab.
      apply gk3c_iff_ndc in E. revert E. apply gk3c_tab_NegImpFreeG, D.
    - apply nd_ndc.
  Qed.

End ssL_nd_ndc.

(** Decidability of GK3c *)

Definition goalc := prod context context.

Instance goalc_eq_dec (gamma delta: goalc) :
  dec (gamma = delta).
Proof.
  unfold dec. repeat decide equality.
Qed.


Section Decidability_GK3c.
  Variable A0 B0 : context.
  Definition Uc := scl (A0 ++ B0).
  Definition Uc_sfc : sf_closed Uc := @scl_closed _.
  Definition Gammac := list_prod (power Uc) (power Uc).

  Definition stepc (Delta : list goalc) (gamma : goalc) : Prop :=
    let (A,B) := gamma in
    (exists u, u el A /\ (u el B \/
       match u with
        | Fal => True
        | Imp s t => (A, rep (s::B) Uc) el Delta /\ (rep (t::A) Uc, B) el Delta
        | And s t => (rep (s::t::A) Uc, B) el Delta
        | Or s t => (rep (s::A) Uc, B) el Delta /\ (rep (t::A) Uc, B) el Delta
        | _ => False
       end))
    \/
    (exists u, u el B /\
       match u with
        | Imp s t => (rep (s::A) Uc, rep (t::B) Uc) el Delta
        | And s t => (A, rep (s::B) Uc) el Delta /\ (A, rep (t::B) Uc) el Delta
        | Or s t => (A, rep (s::t::B) Uc) el Delta
        | _ => False
       end).

  Instance stepc_dec Delta gamma :
    dec (stepc Delta gamma).
  Proof.
    destruct gamma as [A B]; simpl.
    apply or_dec; apply list_exists_dec; intros [x| |s t|s t|s t]; auto.
  Qed.

  Definition Lambdac : list goalc :=
    FCI.C (step := stepc) Gammac.

  Lemma lambdac_closure gamma :
    gamma el Gammac -> stepc Lambdac gamma -> gamma el Lambdac.
  Proof. apply FCI.closure. Qed.

  Lemma lambdac_ind (p : goalc -> Prop) :
    (forall Delta gamma, inclp Delta p ->
      gamma el Gammac -> stepc Delta gamma -> p gamma)
    -> inclp Lambdac p.
  Proof. apply FCI.ind. Qed.

  Lemma gk3c_lambdac A B :
    A <<= Uc -> B <<= Uc -> gk3c A B -> (rep A Uc, rep B Uc) el Lambdac.
  Proof.
    intros D1 D2 D.
    induction D as [A B x E F|A B E|A B s t E F IHF G IHG|A B s t E F IH|
                    A B s t E F IH|A B s t E F IHF G IHG|A B s t E F IHF G IHG|
                    A B s t E F IH].
    - (apply lambdac_closure; [now eauto using in_prod, rep_power|]); simpl.
      left. exists (Var x). auto using rep_in.
    - (apply lambdac_closure; [now eauto using in_prod, rep_power|]); simpl.
      left. exists Fal. auto using rep_in.
    - (apply lambdac_closure; [now eauto using in_prod, rep_power|]); simpl.
      left. exists (Imp s t). split; auto using rep_in. right.
      apply D1 in E. apply Uc_sfc in E as [E1 E2].
      assert (I: (rep A Uc, rep (s :: B) Uc) el Lambdac) by auto.
      assert (J: (rep (t :: A) Uc, rep B Uc) el Lambdac) by auto.
      rewrite rep_eq with (A:=s :: rep B Uc) (B:=s :: B).
      rewrite rep_eq with (A:=t :: rep A Uc) (B:=t :: A); auto.
      + rewrite rep_equi; auto.
      + rewrite rep_equi; auto.
    - (apply lambdac_closure; [now eauto using in_prod, rep_power|]); simpl.
      right. exists (Imp s t). split; auto using rep_in.
      apply D2 in E. apply Uc_sfc in E as [E1 E2].
      assert (I: (rep (s :: A) Uc, rep (t :: B) Uc) el Lambdac) by auto.
      rewrite rep_eq with (B:=s::A).
      rewrite rep_eq with (A:=t :: rep B Uc) (B:=t :: B); auto.
      + setoid_rewrite rep_equi; auto.
      + setoid_rewrite rep_equi; auto.
    - (apply lambdac_closure; [now eauto using in_prod, rep_power|]); simpl.
      left. exists (And s t). split; auto using rep_in. right.
      apply D1 in E. apply Uc_sfc in E as [E1 E2].
      assert (I: (rep (s :: t :: A) Uc, rep B Uc) el Lambdac) by auto.
      rewrite rep_eq with (B:=s :: t :: A); auto.
      rewrite rep_equi; auto.
    - (apply lambdac_closure; [now eauto using in_prod, rep_power|]); simpl.
      right. exists (And s t). split; auto using rep_in.
      apply D2 in E. apply Uc_sfc in E as [E1 E2].
      assert (I: (rep A Uc, rep (s :: B) Uc) el Lambdac) by auto.
      assert (J: (rep A Uc, rep (t :: B) Uc) el Lambdac) by auto.
      rewrite rep_eq with (A:=s :: rep B Uc) (B:=s :: B).
      rewrite rep_eq with (A:=t :: rep B Uc) (B:=t :: B); auto.
      + rewrite rep_equi; auto.
      + rewrite rep_equi; auto.
    - (apply lambdac_closure; [now eauto using in_prod, rep_power|]); simpl.
      left. exists (Or s t). split; auto using rep_in. right.
      apply D1 in E. apply Uc_sfc in E as [E1 E2].
      assert (I: (rep (s :: A) Uc, rep B Uc) el Lambdac) by auto.
      assert (J: (rep (t :: A) Uc, rep B Uc) el Lambdac) by auto.
      rewrite rep_eq with (A:=s :: rep A Uc) (B:=s :: A).
      rewrite rep_eq with (A:=t :: rep A Uc) (B:=t :: A); auto.
      + rewrite rep_equi; auto.
      + rewrite rep_equi; auto.
    - (apply lambdac_closure; [now eauto using in_prod, rep_power|]); simpl.
      right. exists (Or s t). split; auto using rep_in.
      apply D2 in E. apply Uc_sfc in E as [E1 E2].
      assert (I: (rep A Uc, rep (s :: t :: B) Uc) el Lambdac) by auto.
      rewrite rep_eq with (A:=s :: t :: rep B Uc) (B:=s :: t :: B); auto.
      rewrite rep_equi; auto.
  Qed.

  Lemma lambdac_gk3c A B :
    (A,B) el Lambdac -> gk3c A B.
  Proof.
    revert A B.
    cut (inclp Lambdac (fun gamma => gk3c (fst gamma) (snd gamma))).
      { intros E A B. apply E. }
    apply lambdac_ind. intros Delta [A B] E F G; simpl.
    assert (E': forall A B, (A,B) el Delta -> gk3c A B).
      { intros A' B'. apply E. } clear E.
    destruct G as [[[x| |s t|s t|s t] [G [H|H]]]|[[x| |s t|s t|s t] [G H]]];
    try tauto; try (eauto using gk3cW, rep_incl);
    destruct H as [H1 H2]; apply E' in H1; apply E' in H2;
    eauto using gk3cW, rep_incl.
  Qed.

  Theorem gk3c_dec: dec (gk3c A0 B0).
  Proof.
    assert (A1: A0 <<= Uc).
    { apply scl_incl; auto. }
    assert (A2: B0 <<= Uc).
    { apply scl_incl; auto. }
    apply dec_prop_iff with (X:= gk3c (rep A0 Uc) (rep B0 Uc)).
    - split; intros E; apply (gk3cW E), rep_equi; destruct (rep_equi A1); auto.
    - decide ((rep A0 Uc, rep B0 Uc) el Lambdac) as [E|E].
      + left. apply lambdac_gk3c, E.
      + right. intros F. apply E.
        apply gk3c_lambdac; auto.
        apply (gk3cW F); auto; apply rep_equi; auto.
  Qed.

End Decidability_GK3c.

Theorem ndc_dec_gk3c A s: dec (ndc A s).
Proof. apply (dec_prop_iff (gk3c_iff_ndc A s)). apply gk3c_dec. Qed.


Lemma gk3c_bentG A B:
  gk3c A B -> bent (A ++ neg B) Fal.
Proof.
  intros E. induction E as [A B x E F|A B E|A B s t E _ IHF _ IHG|
                            A B s t E _ IH|A B s t E _ IH|
                            A B s t E _ IHF _ IHG|A B s t E _ IHF _ IHG|
                            A B s t E _ IH];
  intros alpha D; auto.
  - assert (satis alpha (Var x)) by auto.
    assert (satis alpha (Not (Var x))) by auto using neg_in. auto.
  - assert (H: satis alpha (Imp s t)) by auto.
    apply IHF. intros u F. apply in_app_iff in F as [F|[F|F]]; subst; auto.
    intros F. apply (IHG alpha). intros u G.
    apply in_app_iff in G as [[G|G]|G]; subst; auto.
  - assert (H: satis alpha (Not (Imp s t))) by auto using neg_in.
    apply H. intros K. exfalso. apply (IH alpha). intros u F.
    apply in_app_iff in F as [[F|F]|[F|F]]; subst; auto.
    intros F. apply H. intros _. auto.
  - assert (H: satis alpha (And s t)) by auto. destruct H as [H K].
    apply IH. intros u F. apply in_app_iff in F as [[F|[F|F]]|F]; subst; auto.
  - assert (H: satis alpha (Not (And s t))) by auto using neg_in.
    apply IHF. intros u F. apply in_app_iff in F as [F|[F|F]]; subst; auto.
    intros F. apply (IHG alpha). intros u G.
    apply in_app_iff in G as [G|[G|G]]; subst; auto.
    intros G. apply H. simpl. auto.
  - assert (H: satis alpha (Or s t)) by auto. destruct H as [H|H].
    + apply IHF. intros u F. apply in_app_iff in F as [[F|F]|F]; subst; auto.
    + apply IHG. intros u G. apply in_app_iff in G as [[G|G]|G]; subst; auto.
  - assert (H: satis alpha (Not (Or s t))) by auto using neg_in.
    apply IH. intros u F. apply in_app_iff in F as [F|[F|[F|F]]];
    subst; auto; intros K; apply H; simpl; auto.
Qed.

Lemma gk3c_bent A s :
  gk3c A [s] -> bent A s.
Proof.
  intros E. apply gk3c_bentG in E.
  intros alpha F. decide (satis alpha s) as [H|H]; auto. exfalso.
  apply (E alpha). intros t K. apply in_app_iff in K as [K|[K|K]]; subst; auto.
  exfalso. auto.
Qed.

Lemma ref_gk3c:
  ref_pred (fun C => gk3c (map uns C) []).
Proof.
  split; simpl.
  - intros C E. apply gk3cF. apply (in_map uns _ _ E).
  - intros C C' E F. apply (gk3cLW F); auto. apply incl_map, E.
  - intros x C E F. apply gk3cIL with (s:=Var x) (t:=Fal); auto.
    + apply (in_map uns _ _ F).
    + apply gk3cA with (s:=Var x); auto. apply (in_map uns _ _ E).
  - intros s t C E F. apply gk3cFL_iff in E.
    apply gk3cIL with (s:=s) (t:=t); eauto using gk3cLW.
  - intros s t C E. apply gk3cFL_iff. apply gk3cIR with (s:=s) (t:=t); auto.
    apply gk3cFL_iff. apply (gk3cW E); auto.
  - intros s t C E. apply gk3cAL with (s:=s) (t:=t);
    eauto using gk3cLW, incl_shift.
  - intros s t C E F. apply gk3cFL_iff in E. apply gk3cFL_iff in F.
    apply gk3cFL_iff. apply gk3cAR with (s:=s) (t:=t); eauto using gk3cRW.
  - intros s t C E F. apply gk3cOL with (s:=s) (t:=t); eauto using gk3cLW.
  - intros s t C E. repeat apply gk3cFL_inv in E. apply gk3cFL_iff.
    apply gk3cOR with (s:=s) (t:=t); eauto using gk3cRW.
  - intros C E [alpha F]. apply gk3c_bentG in E. apply (E alpha).
    intros t H. apply in_app_iff in H as [H|H].
    + apply in_map_iff in H as [u [H K]]. subst. revert F K. apply satis_in.
    + exfalso. auto.
Qed.


Lemma gk3c_refute A B:
  gk3c A B <-> gk3c (A ++ neg B) [].
Proof.
  revert A. induction B as [|s B]; intros A; simpl.
  - rewrite app_nil_r. tauto.
  - split; intros E.
    + apply gk3cFL in E. apply IHB in E. apply (gk3cLW E); auto.
    + assert (H: gk3c ((Not s :: A) ++ neg B) []) by (apply (gk3cLW E); auto).
      apply IHB in H. apply gk3cFL_iff. auto.
Qed.

Lemma uns_neg B :
  map uns (map Neg B) = neg B.
Proof. induction B; simpl; congruence. Qed.

Lemma gk3c_dec_ref A B: dec (gk3c A B).
Proof.
  generalize (ref_dec ref_gk3c (map Pos A ++ map Neg B)).
  simpl. rewrite map_app. rewrite uns_pos. rewrite uns_neg.
  apply dec_prop_iff.
  symmetry. apply gk3c_refute.
Qed.

(** * Agreement of classical Gentzen with boolean entailment *)
Lemma gk3c_iff_sem A s :
  gk3c A [s] <-> bent A s.
Proof.
  rewrite bent_iff_unsat.
  rewrite gk3c_refute.
  generalize (ref_iff_unsat ref_gk3c (-s :: map Pos A)).
  simpl. rewrite uns_pos. intros C. split; intros E.
  - apply C. apply (gk3cLW E); auto.
  - apply C in E. apply (gk3cLW E); auto.
Qed.

(** * Classical Hilbert System Hc *)

Inductive hilc (A : context): form -> Prop :=
  | hilcK s t:     hilc A (FK s t)
  | hilcS s t u:   hilc A (FS s t u)
  | hilcAL s t:    hilc A (Imp (And s t) s)
  | hilcAR s t:    hilc A (Imp (And s t) t)
  | hilcAI s t:    hilc A (Imp s (Imp t (And s t)))
  | hilcOL s t:    hilc A (Imp s (Or s t))
  | hilcOR s t:    hilc A (Imp t (Or s t))
  | hilcOE s t u:  hilc A (Imp (Imp s u) (Imp (Imp t u) (Imp (Or s t) u)))
  | hilcDN s:      hilc A (Imp (Not (Not s)) s)
  | hilcA s:       s el A -> hilc A s
  | hilcMP s t:    hilc A (Imp s t) -> hilc A s -> hilc A t.

(** Equivalence of Nc and Hc *)
Lemma hilc_ndc A s :
  hilc A s -> ndc A s.

Proof.
  intros E. induction E as [s t|s t u|s t|s t|s t|s t|s t|
                            s t u|s|s E|s t E1 IHE1 E2 IHE2]; auto.
  - apply ndcII, ndcII, ndcA. auto.
  - repeat apply ndcII.
    apply ndcIE with (s:= Imp t u).
    + apply ndcII. apply ndcIE with (s:=t); auto.
      apply ndcIE with (s:=s); auto.
    + apply ndcIE with (s:=s); auto.
  - apply ndcII. apply ndcAE with (s:=s) (t:=t); auto.
  - apply ndcII. apply ndcAE with (s:=s) (t:=t); auto. 
  - apply ndcII, ndcII. apply ndcAI; auto.
  - repeat apply ndcII. apply ndcOE with (s:=s) (t:=t); auto.
    + apply ndcIE with (s:=s); auto.
    + apply ndcIE with (s:=t); auto.
  - apply ndcII, ndcC. apply ndcIE with (s:=Not s) (t:=Fal); auto.
  - apply ndcIE with (s:=s) (t:=t); auto.
Qed.

Lemma hilcAK A s t:
  hilc A s -> hilc A (Imp t s).
Proof. apply hilcMP. apply hilcK. Qed.

Lemma hilcAS A s t u :
  hilc A (Imp s (Imp t u)) -> hilc A (Imp s t) -> hilc A (Imp s u).
Proof. intros B. apply hilcMP. revert B. apply hilcMP. apply hilcS. Qed.

Lemma hilcI A s: hilc A (Imp s s).
Proof.
  assert (E:= hilcS A s (Imp s s) s).
  assert (F:= hilcK A s (Imp s s)).
  assert (G:= hilcK A s s).
  unfold FS, FK in *.
  exact (hilcMP (hilcMP E F) G).
Qed.

Lemma hilcD w A v :
  hilc (w::A) v -> hilc A (Imp w v).
Proof.
  intros E. induction E as [s t|s t u|s t|s t|s t|s t|s t|
                            s t u|s|s E|s t E1 IHE1 E2 IHE2]; auto.
  - apply hilcAK, hilcK.
  - apply hilcAK, hilcS.
  - apply hilcAK, hilcAL.
  - apply hilcAK, hilcAR.
  - apply hilcAK, hilcAI.
  - apply hilcAK, hilcOL.
  - apply hilcAK, hilcOR.
  - apply hilcAK, hilcOE.
  - apply hilcAK, hilcDN.
  -  destruct E as [E|E].
    + subst. apply hilcI.
    + apply hilcAK. apply hilcA, E.
  - apply hilcAS with (t:=s); auto.
Qed.

Lemma ndc_hilc A s :
  ndc A s -> hilc A s.
Proof. 
  intros E. induction E as [A s E|A s E IH|A s t E IH|A s t E IHE F IHF|
                            A s t E IHE F IHF|A s t u E IHE F IHF|A s t E IH|
                            A s t E IH|A s t u E IHE F IHF G IHG]; auto.
  - apply hilcA, E.
  - apply hilcMP with (s:=Not (Not s)).
    + apply hilcDN.
    + apply hilcD, IH.
  - apply hilcD, IH.
  - exact (hilcMP IHE IHF).
  - exact (hilcMP (hilcMP (hilcAI A s t) IHE) IHF).
  - apply hilcMP with (s:=s).
    + apply hilcMP with (s:=t).
      * apply hilcD, hilcD. auto.
      * exact (hilcMP (hilcAR A s t) IHE).
    + exact (hilcMP (hilcAL A s t) IHE).
  - exact (hilcMP (hilcOL A s t) IH).
  - exact (hilcMP (hilcOR A s t) IH).
  - apply hilcMP with (s:=Or s t); auto.
    apply hilcMP with (s:=Imp t u).
    + apply hilcMP with (s:=Imp s u).
      * apply hilcOE.
      * apply hilcD. exact IHF.
    + apply hilcD. exact IHG.
Qed.

Theorem hilc_iff_ndc A s:
  hilc A s <-> ndc A s.
Proof. split. apply hilc_ndc. apply ndc_hilc. Qed.